use pretty_assertions::assert_eq;
use snowman_connector::{
    query::DatabaseSchema,
    schema::{Column, Table},
};
use snowman_generator::generate_schema_python_code;

#[tokio::test]
async fn test_generate_schema_python_code() {
    let database_schema = DatabaseSchema {
        database_name: "DATABASE".to_string(),
        schema_name: "SCHEMA".to_string(),
    };
    let tables = vec![Table {
        database_name: database_schema.database_name,
        schema_name: database_schema.schema_name,
        table_name: "USER".to_string(),
        comment: Some("User Table".to_string()),
        columns: vec![
            Column {
                column_name: "ID".to_string(),
                data_type: "INTEGER".to_string(),
                is_nullable: false,
                comment: Some("User ID".to_string()),
                default_value: None,
            },
            Column {
                column_name: "NAME".to_string(),
                data_type: "TEXT".to_string(),
                is_nullable: false,
                comment: Some("User Name".to_string()),
                default_value: None,
            },
            Column {
                column_name: "CREATED_AT".to_string(),
                data_type: "TIMESTAMP".to_string(),
                is_nullable: false,
                comment: Some("Created At".to_string()),
                default_value: Some("CURRENT_TIMESTAMP()".to_string()),
            },
        ],
    }];

    let code = generate_schema_python_code(
        &tables,
        &Default::default(),
        &Default::default(),
        &Default::default(),
    )
    .await
    .unwrap();

    assert_eq!(
        code,
        r#"
#
# Code generated by snowman; DO NOT EDIT.
#
# For more information about snowman,
# please refer to https://github.com/yassun7010/snowman-py .
#

import typing
import snowman
import pydantic

if typing.TYPE_CHECKING:
    class _UserInsertTypedDict(typing.TypedDict):
        id: snowman.datatype.INTEGER
        """User ID"""

        name: snowman.datatype.TEXT
        """User Name"""

        created_at: typing.NotRequired[snowman.datatype.TIMESTAMP]
        """Created At"""

    class _UserUpdateTypedDict(typing.TypedDict):
        id: typing.NotRequired[snowman.datatype.INTEGER]
        """User ID"""

        name: typing.NotRequired[snowman.datatype.TEXT]
        """User Name"""

        created_at: typing.NotRequired[snowman.datatype.TIMESTAMP]
        """Created At"""

# TABLE: DATABASE.SCHEMA.USER
@snowman.table("DATABASE", "SCHEMA", "USER")
class User(pydantic.BaseModel, snowman.Table["_UserInsertTypedDict","_UserUpdateTypedDict",]):
    """User Table"""
    model_config = pydantic.ConfigDict(populate_by_name=True)

    id: typing.Annotated[snowman.datatype.INTEGER, pydantic.Field(title="User ID", alias="ID"),]
    """User ID"""

    name: typing.Annotated[snowman.datatype.TEXT, pydantic.Field(title="User Name", alias="NAME"),]
    """User Name"""

    created_at: snowman.datatype.TIMESTAMP = pydantic.Field(title="Created At", alias="CREATED_AT", default_factory=snowman.datatype.TIMESTAMP.now)
    """Created At"""
"#
        .strip_prefix('\n')
        .unwrap()
    )
}

#[tokio::test]
async fn test_generate_schema_python_code_of_empty_columns_table() {
    let database_schema = DatabaseSchema {
        database_name: "DATABASE".to_string(),
        schema_name: "SCHEMA".to_string(),
    };
    let tables = vec![Table {
        database_name: database_schema.database_name,
        schema_name: database_schema.schema_name,
        table_name: "USER".to_string(),
        comment: Some("User Table".to_string()),
        columns: vec![],
    }];

    let code = generate_schema_python_code(
        &tables,
        &Default::default(),
        &Default::default(),
        &Default::default(),
    )
    .await
    .unwrap();

    assert_eq!(
        code,
        r#"
#
# Code generated by snowman; DO NOT EDIT.
#
# For more information about snowman,
# please refer to https://github.com/yassun7010/snowman-py .
#

import typing
import snowman
import pydantic

if typing.TYPE_CHECKING:
    class _UserInsertTypedDict(typing.TypedDict):
        pass

    class _UserUpdateTypedDict(typing.TypedDict):
        pass

# TABLE: DATABASE.SCHEMA.USER
@snowman.table("DATABASE", "SCHEMA", "USER")
class User(pydantic.BaseModel, snowman.Table["_UserInsertTypedDict","_UserUpdateTypedDict",]):
    """User Table"""
    model_config = pydantic.ConfigDict(populate_by_name=True)
"#
        .strip_prefix('\n')
        .unwrap()
    )
}
